"""Tests for the transfer APIs using liquid classes."""

import mock
import pytest

from opentrons.protocol_api import ProtocolContext
from opentrons.protocol_api.core.engine import InstrumentCore
from opentrons.protocol_api.core.engine.transfer_components_executor import (
    LiquidAndAirGapPair,
    TransferType,
)
from opentrons.protocols.advanced_control.transfers.common import (
    TransferTipPolicyV2Type,
)
from opentrons.types import Location, Point


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_water_transfer_with_volume_more_than_tip_max(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should run the transfer steps without any errors.

    This test only checks that various supported configurations for a transfer
    analyze successfully. It doesn't check whether the steps are as expected.
    That will be covered in analysis snapshot tests.
    """
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    pipette_1k = simulated_protocol_context.load_instrument(
        "flex_1channel_1000", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    with mock.patch.object(
        InstrumentCore,
        "pick_up_tip",
        side_effect=InstrumentCore.pick_up_tip,
        autospec=True,
    ) as patched_pick_up_tip:
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")

        pipette_1k.transfer_with_liquid_class(
            liquid_class=water,
            volume=60,
            source=nest_plate.rows()[0],
            dest=arma_plate.rows()[0],
            new_tip="always",
            trash_location=trash,
        )
        assert patched_pick_up_tip.call_count == 24
        patched_pick_up_tip.reset_mock()

        pipette_1k.transfer_with_liquid_class(
            liquid_class=water,
            volume=100,
            source=nest_plate.rows()[0],
            dest=arma_plate.rows()[0],
            new_tip="per source",
            trash_location=trash,
        )
        assert patched_pick_up_tip.call_count == 12
        patched_pick_up_tip.reset_mock()

        pipette_1k.pick_up_tip()
        pipette_1k.transfer_with_liquid_class(
            liquid_class=water,
            volume=50,
            source=nest_plate.rows()[0],
            dest=arma_plate.rows()[0],
            new_tip="never",
            trash_location=trash,
        )
        pipette_1k.drop_tip()
        assert patched_pick_up_tip.call_count == 1


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_order_of_water_transfer_steps(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should run the transfer steps without any errors.

    This test only checks that various supported configurations for a transfer
    analyze successfully. It doesn't check whether the steps are as expected.
    That will be covered in analysis snapshot tests.
    """
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    pipette_50 = simulated_protocol_context.load_instrument(
        "flex_1channel_50", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    with (
        mock.patch.object(
            InstrumentCore,
            "load_liquid_class",
            side_effect=InstrumentCore.load_liquid_class,
            autospec=True,
        ) as patched_load_liquid_class,
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class",
            side_effect=InstrumentCore.dispense_liquid_class,
            autospec=True,
        ) as patched_dispense,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_load_liquid_class, "load_liquid_class")
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(patched_dispense, "dispense_liquid_class")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_50.transfer_with_liquid_class(
            liquid_class=water,
            volume=40,
            source=nest_plate.rows()[0][:2],
            dest=arma_plate.rows()[0][:2],
            new_tip="always",
            trash_location=trash,
        )
        expected_calls = [
            mock.call.load_liquid_class(
                mock.ANY,
                name="water",
                transfer_properties=mock.ANY,
                tiprack_uri="opentrons/opentrons_flex_96_tiprack_50ul/1",
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=40,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=40,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=40,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=40, air_gap=0.1)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=40,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=40,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=40,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=40, air_gap=0.1)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
        ]
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_order_of_water_transfer_steps_with_new_tip_per_destination(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should run the transfer steps while picking up a new tip only for a new destination."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    pipette_50 = simulated_protocol_context.load_instrument(
        "flex_1channel_50", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    with (
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class",
            side_effect=InstrumentCore.dispense_liquid_class,
            autospec=True,
        ) as patched_dispense,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(patched_dispense, "dispense_liquid_class")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_50.transfer_with_liquid_class(
            liquid_class=water,
            volume=60,
            source=nest_plate.rows()[0][:2],
            dest=nest_plate.rows()[1][:2],
            new_tip="per destination",
            trash_location=trash,
        )
        expected_calls_per_tip = [
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=30,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=30,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=30,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=30, air_gap=0.1)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=30,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0.1)],
                volume_for_pipette_mode_configuration=30,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=30,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=30, air_gap=0.1)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
        ]
        assert (
            mock_manager.mock_calls == expected_calls_per_tip + expected_calls_per_tip
        )


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_order_of_water_transfer_steps_with_return_tip(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should run the transfer steps without any errors and return tips.

    This test only checks that various supported configurations for a transfer
    analyze successfully. It doesn't check whether the steps are as expected.
    That will be covered in analysis snapshot tests.
    """
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    pipette_50 = simulated_protocol_context.load_instrument(
        "flex_1channel_50", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    with (
        mock.patch.object(
            InstrumentCore,
            "load_liquid_class",
            side_effect=InstrumentCore.load_liquid_class,
            autospec=True,
        ) as patched_load_liquid_class,
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class",
            side_effect=InstrumentCore.dispense_liquid_class,
            autospec=True,
        ) as patched_dispense,
        mock.patch.object(
            InstrumentCore,
            "drop_tip",
            side_effect=InstrumentCore.drop_tip,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_load_liquid_class, "load_liquid_class")
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(patched_dispense, "dispense_liquid_class")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip")
        pipette_50.transfer_with_liquid_class(
            liquid_class=water,
            volume=40,
            source=nest_plate.rows()[0][:2],
            dest=arma_plate.rows()[0][:2],
            new_tip="always",
            trash_location=trash,
            return_tip=True,
        )
        expected_calls = [
            mock.call.load_liquid_class(
                mock.ANY,
                name="water",
                transfer_properties=mock.ANY,
                tiprack_uri="opentrons/opentrons_flex_96_tiprack_50ul/1",
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=40,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=40,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=40,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=40, air_gap=0.1)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.drop_tip(
                mock.ANY,
                location=None,
                well_core=mock.ANY,
                home_after=False,
                alternate_drop_location=False,
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=40,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=40,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=40,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=40, air_gap=0.1)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.drop_tip(
                mock.ANY,
                location=None,
                well_core=mock.ANY,
                home_after=False,
                alternate_drop_location=False,
            ),
        ]
        assert len(mock_manager.mock_calls) == 9
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_order_of_water_transfer_steps_with_no_new_tips(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should run the transfer steps without any errors.

    This test only checks that various supported configurations for a transfer
    analyze successfully. It doesn't check whether the steps are as expected.
    That will be covered in analysis snapshot tests.
    """
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    pipette_50 = simulated_protocol_context.load_instrument(
        "flex_1channel_50", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    pipette_50.pick_up_tip()
    with (
        mock.patch.object(
            InstrumentCore,
            "load_liquid_class",
            side_effect=InstrumentCore.load_liquid_class,
            autospec=True,
        ) as patched_load_liquid_class,
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class",
            side_effect=InstrumentCore.dispense_liquid_class,
            autospec=True,
        ) as patched_dispense,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_load_liquid_class, "load_liquid_class")
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(patched_dispense, "dispense_liquid_class")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_50.transfer_with_liquid_class(
            liquid_class=water,
            volume=40,
            source=nest_plate.rows()[0][:2],
            dest=arma_plate.rows()[0][:2],
            new_tip="never",
            trash_location=trash,
        )
        expected_calls = [
            mock.call.load_liquid_class(
                mock.ANY,
                name="water",
                transfer_properties=mock.ANY,
                tiprack_uri="opentrons/opentrons_flex_96_tiprack_50ul/1",
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=40,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=40,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=40,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=40, air_gap=0.1)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=40,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0.1)],
                volume_for_pipette_mode_configuration=40,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=40,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=40, air_gap=0.1)],
                add_final_air_gap=False,
                trash_location=mock.ANY,
            ),
        ]
        assert len(mock_manager.mock_calls) == len(expected_calls)
        assert mock_manager.mock_calls[2] == expected_calls[2]


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
@pytest.mark.parametrize("new_tip", ["once", "always"])
def test_order_of_water_consolidate_steps(
    simulated_protocol_context: ProtocolContext,
    new_tip: TransferTipPolicyV2Type,
) -> None:
    """It should run the consolidate steps without any errors.

    This test only checks that various supported configurations for a consolidation
    analyze successfully. It doesn't check whether the steps are as expected.
    That will be covered in analysis snapshot tests.
    """
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    pipette_50 = simulated_protocol_context.load_instrument(
        "flex_1channel_50", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    with (
        mock.patch.object(
            InstrumentCore,
            "load_liquid_class",
            side_effect=InstrumentCore.load_liquid_class,
            autospec=True,
        ) as patched_load_liquid_class,
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class",
            side_effect=InstrumentCore.dispense_liquid_class,
            autospec=True,
        ) as patched_dispense,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_load_liquid_class, "load_liquid_class")
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(patched_dispense, "dispense_liquid_class")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_50.consolidate_with_liquid_class(
            liquid_class=water,
            volume=25,
            source=nest_plate.rows()[0][:2],
            dest=arma_plate.wells()[0],
            new_tip=new_tip,
            trash_location=trash,
        )
        expected_calls = [
            mock.call.load_liquid_class(
                mock.ANY,
                name="water",
                transfer_properties=mock.ANY,
                tiprack_uri="opentrons/opentrons_flex_96_tiprack_50ul/1",
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=25,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=50,
                current_volume=0.0,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=25,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=25, air_gap=0.1)],
                volume_for_pipette_mode_configuration=None,
                current_volume=25.0,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=50,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=50, air_gap=0)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
        ]
        assert len(mock_manager.mock_calls) == 6
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_order_of_water_consolidate_steps_larger_volume_than_tip(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should run the consolidate steps without any errors.

    This test only checks that various supported configurations for a consolidation
    analyze successfully. It doesn't check whether the steps are as expected.
    That will be covered in analysis snapshot tests.
    """
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    pipette_50 = simulated_protocol_context.load_instrument(
        "flex_1channel_50", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    with (
        mock.patch.object(
            InstrumentCore,
            "load_liquid_class",
            side_effect=InstrumentCore.load_liquid_class,
            autospec=True,
        ) as patched_load_liquid_class,
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class",
            side_effect=InstrumentCore.dispense_liquid_class,
            autospec=True,
        ) as patched_dispense,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_load_liquid_class, "load_liquid_class")
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(patched_dispense, "dispense_liquid_class")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_50.consolidate_with_liquid_class(
            liquid_class=water,
            volume=30,
            source=nest_plate.rows()[0][:2],
            dest=arma_plate.wells()[0],
            new_tip="once",
            trash_location=trash,
        )
        expected_calls = [
            mock.call.load_liquid_class(
                mock.ANY,
                name="water",
                transfer_properties=mock.ANY,
                tiprack_uri="opentrons/opentrons_flex_96_tiprack_50ul/1",
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=30,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=30,
                current_volume=0.0,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=30,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=30, air_gap=0.1)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=30,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0.1)],
                volume_for_pipette_mode_configuration=30,
                current_volume=0.0,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=30,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=30, air_gap=0.1)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
        ]
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_order_of_water_consolidate_steps_larger_volume_than_tip_with_new_tip_always(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should pick up new tips & probe liquid before every group of aspirates."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    pipette_50 = simulated_protocol_context.load_instrument(
        "flex_1channel_50",
        mount="left",
        tip_racks=[tiprack],
        liquid_presence_detection=True,
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    with (
        mock.patch.object(
            InstrumentCore,
            "load_liquid_class",
            side_effect=InstrumentCore.load_liquid_class,
            autospec=True,
        ) as patched_load_liquid_class,
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class",
            side_effect=InstrumentCore.dispense_liquid_class,
            autospec=True,
        ) as patched_dispense,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_load_liquid_class, "load_liquid_class")
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(patched_dispense, "dispense_liquid_class")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_50.consolidate_with_liquid_class(
            liquid_class=water,
            volume=30,
            source=nest_plate.rows()[0][:2],
            dest=arma_plate.wells()[0],
            new_tip="always",
            trash_location=trash,
        )
        expected_calls = [
            mock.call.load_liquid_class(
                mock.ANY,
                name="water",
                transfer_properties=mock.ANY,
                tiprack_uri="opentrons/opentrons_flex_96_tiprack_50ul/1",
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=30,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=30.0,
                current_volume=0.0,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=30,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=30, air_gap=0.1)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=30,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=30.0,
                current_volume=0.0,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=30,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=30, air_gap=0.1)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
        ]
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_order_of_water_consolidate_steps_with_no_new_tips(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should run the consolidate steps without any errors.

    This test only checks that various supported configurations for a consolidation
    analyze successfully. It doesn't check whether the steps are as expected.
    That will be covered in analysis snapshot tests.
    """
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    pipette_50 = simulated_protocol_context.load_instrument(
        "flex_1channel_50", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    pipette_50.pick_up_tip()
    with (
        mock.patch.object(
            InstrumentCore,
            "load_liquid_class",
            side_effect=InstrumentCore.load_liquid_class,
            autospec=True,
        ) as patched_load_liquid_class,
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class",
            side_effect=InstrumentCore.dispense_liquid_class,
            autospec=True,
        ) as patched_dispense,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_load_liquid_class, "load_liquid_class")
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(patched_dispense, "dispense_liquid_class")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_50.consolidate_with_liquid_class(
            liquid_class=water,
            volume=25,
            source=nest_plate.rows()[0][:2],
            dest=arma_plate.wells()[0],
            new_tip="never",
            trash_location=trash,
        )
        expected_calls = [
            mock.call.load_liquid_class(
                mock.ANY,
                name="water",
                transfer_properties=mock.ANY,
                tiprack_uri="opentrons/opentrons_flex_96_tiprack_50ul/1",
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=25,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=50,
                current_volume=0.0,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=25,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=25, air_gap=0.1)],
                volume_for_pipette_mode_configuration=None,
                current_volume=25.0,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=50,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=50, air_gap=0)],
                add_final_air_gap=False,
                trash_location=mock.ANY,
            ),
        ]
        assert len(mock_manager.mock_calls) == 4
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_order_of_water_consolidate_steps_with_return_tip(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should run the consolidate steps without any errors and return tips.

    This test only checks that various supported configurations for a consolidation
    analyze successfully. It doesn't check whether the steps are as expected.
    That will be covered in analysis snapshot tests.
    """
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    pipette_50 = simulated_protocol_context.load_instrument(
        "flex_1channel_50", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    with (
        mock.patch.object(
            InstrumentCore,
            "load_liquid_class",
            side_effect=InstrumentCore.load_liquid_class,
            autospec=True,
        ) as patched_load_liquid_class,
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class",
            side_effect=InstrumentCore.dispense_liquid_class,
            autospec=True,
        ) as patched_dispense,
        mock.patch.object(
            InstrumentCore,
            "drop_tip",
            side_effect=InstrumentCore.drop_tip,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_load_liquid_class, "load_liquid_class")
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(patched_dispense, "dispense_liquid_class")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip")
        pipette_50.consolidate_with_liquid_class(
            liquid_class=water,
            volume=25,
            source=nest_plate.rows()[0][:2],
            dest=arma_plate.wells()[0],
            new_tip="once",
            trash_location=trash,
            return_tip=True,
        )
        expected_calls = [
            mock.call.load_liquid_class(
                mock.ANY,
                name="water",
                transfer_properties=mock.ANY,
                tiprack_uri="opentrons/opentrons_flex_96_tiprack_50ul/1",
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=25,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=50,
                current_volume=0.0,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=25,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=25, air_gap=0.1)],
                volume_for_pipette_mode_configuration=None,
                current_volume=25.0,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=50,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.MANY_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=50, air_gap=0)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.drop_tip(
                mock.ANY,
                location=None,
                well_core=mock.ANY,
                home_after=False,
                alternate_drop_location=False,
            ),
        ]
        assert len(mock_manager.mock_calls) == 6
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_water_distribution_with_volume_more_than_tip_max(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should execute the distribute steps with the expected tip pick ups."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_200ul", "D1"
    )
    pipette_1k = simulated_protocol_context.load_instrument(
        "flex_1channel_1000", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    water_props = water.get_for(pipette_1k, tiprack)
    water_props.multi_dispense.retract.blowout.location = "destination"  # type: ignore[union-attr]
    water_props.multi_dispense.retract.blowout.flow_rate = pipette_1k.flow_rate.blow_out  # type: ignore[union-attr]
    water_props.multi_dispense.retract.blowout.enabled = True  # type: ignore[union-attr]
    with mock.patch.object(
        InstrumentCore,
        "pick_up_tip",
        side_effect=InstrumentCore.pick_up_tip,
        autospec=True,
    ) as patched_pick_up_tip:
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")

        pipette_1k.distribute_with_liquid_class(
            liquid_class=water,
            volume=100,
            source=nest_plate.rows()[0][0],
            dest=arma_plate.rows()[0],
            new_tip="once",
            trash_location=trash,
        )
        assert patched_pick_up_tip.call_count == 1
        patched_pick_up_tip.reset_mock()

        pipette_1k.pick_up_tip()
        pipette_1k.distribute_with_liquid_class(
            liquid_class=water,
            volume=50,
            source=nest_plate.rows()[0][0],
            dest=arma_plate.rows()[0],
            new_tip="never",
            trash_location=trash,
        )
        pipette_1k.drop_tip()
        assert patched_pick_up_tip.call_count == 1
        patched_pick_up_tip.reset_mock()

        pipette_1k.distribute_with_liquid_class(
            liquid_class=water,
            volume=50,
            source=nest_plate.rows()[0][0],
            dest=arma_plate.rows()[0],
            new_tip="always",
            trash_location=trash,
        )
        assert patched_pick_up_tip.call_count == 3


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_order_of_water_distribution_steps_using_multi_dispense(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should distribute the liquid using multi-dispense steps in the expected order."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_1000ul", "D1"
    )
    pipette_1k = simulated_protocol_context.load_instrument(
        "flex_1channel_1000", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    water_props = water.get_for(pipette_1k, tiprack)

    assert water_props.multi_dispense is not None
    water_props.multi_dispense.retract.blowout.location = "destination"
    water_props.multi_dispense.retract.blowout.flow_rate = pipette_1k.flow_rate.blow_out
    water_props.multi_dispense.retract.blowout.enabled = True

    expected_conditioning_volume = 4.5
    expected_disposal_volume = 5.5
    water_props.multi_dispense.conditioning_by_volume.set_for_all_volumes(
        expected_conditioning_volume
    )
    water_props.multi_dispense.disposal_by_volume.set_for_all_volumes(
        expected_disposal_volume
    )

    with (
        mock.patch.object(
            InstrumentCore,
            "load_liquid_class",
            side_effect=InstrumentCore.load_liquid_class,
            autospec=True,
        ) as patched_load_liquid_class,
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class_during_multi_dispense",
            side_effect=InstrumentCore.dispense_liquid_class_during_multi_dispense,
            autospec=True,
        ) as patched_dispense,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_load_liquid_class, "load_liquid_class")
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(
            patched_dispense, "dispense_liquid_class_during_multi_dispense"
        )
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_1k.distribute_with_liquid_class(
            liquid_class=water,
            volume=400,
            source=nest_plate.rows()[0][1],
            dest=arma_plate.rows()[0][:4],
            new_tip="once",
            trash_location=trash,
        )
        expected_calls = [
            mock.call.load_liquid_class(
                mock.ANY,
                name="water",
                transfer_properties=mock.ANY,
                tiprack_uri="opentrons/opentrons_flex_96_tiprack_1000ul/1",
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=800 + expected_conditioning_volume + expected_disposal_volume,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                conditioning_volume=expected_conditioning_volume,
                volume_for_pipette_mode_configuration=400,
            ),
            mock.call.dispense_liquid_class_during_multi_dispense(
                mock.ANY,
                volume=400,
                dest=(
                    Location(Point(), arma_plate.rows()[0][0]),
                    arma_plate.rows()[0][0]._core,
                ),
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[
                    LiquidAndAirGapPair(
                        liquid=800 + expected_disposal_volume, air_gap=0
                    )
                ],
                add_final_air_gap=True,
                trash_location=mock.ANY,
                conditioning_volume=expected_conditioning_volume,
                disposal_volume=expected_disposal_volume,
                is_last_dispense_in_tip=False,
            ),
            mock.call.dispense_liquid_class_during_multi_dispense(
                mock.ANY,
                volume=400,
                dest=(
                    Location(Point(), arma_plate.rows()[0][1]),
                    arma_plate.rows()[0][1]._core,
                ),
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[
                    LiquidAndAirGapPair(
                        liquid=400 + expected_disposal_volume, air_gap=0
                    )
                ],
                add_final_air_gap=True,
                trash_location=mock.ANY,
                conditioning_volume=expected_conditioning_volume,
                disposal_volume=expected_disposal_volume,
                is_last_dispense_in_tip=True,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=800 + expected_conditioning_volume + expected_disposal_volume,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=10)],
                conditioning_volume=expected_conditioning_volume,
                volume_for_pipette_mode_configuration=400,
            ),
            mock.call.dispense_liquid_class_during_multi_dispense(
                mock.ANY,
                volume=400,
                dest=(
                    Location(Point(), arma_plate.rows()[0][2]),
                    arma_plate.rows()[0][2]._core,
                ),
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[
                    LiquidAndAirGapPair(
                        liquid=800 + expected_disposal_volume, air_gap=0
                    )
                ],
                add_final_air_gap=True,
                trash_location=mock.ANY,
                conditioning_volume=expected_conditioning_volume,
                disposal_volume=expected_disposal_volume,
                is_last_dispense_in_tip=False,
            ),
            mock.call.dispense_liquid_class_during_multi_dispense(
                mock.ANY,
                volume=400,
                dest=(
                    Location(Point(), arma_plate.rows()[0][3]),
                    arma_plate.rows()[0][3]._core,
                ),
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[
                    LiquidAndAirGapPair(
                        liquid=400 + expected_disposal_volume, air_gap=0
                    )
                ],
                add_final_air_gap=True,
                trash_location=mock.ANY,
                conditioning_volume=expected_conditioning_volume,
                disposal_volume=expected_disposal_volume,
                is_last_dispense_in_tip=True,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
        ]
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_order_of_water_distribution_steps_using_multi_dispense_without_conditioning_volume(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should distribute the liquid using multi-dispense steps in the expected order.

    It should add air gaps between dispenses due to no conditioning volume.
    """
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_1000ul", "D1"
    )
    pipette_1k = simulated_protocol_context.load_instrument(
        "flex_1channel_1000", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    water_props = water.get_for(pipette_1k, tiprack)

    assert water_props.multi_dispense is not None
    water_props.multi_dispense.retract.blowout.location = "destination"
    water_props.multi_dispense.retract.blowout.flow_rate = pipette_1k.flow_rate.blow_out
    water_props.multi_dispense.retract.blowout.enabled = True
    disposal_volume = 5.5
    water_props.multi_dispense.conditioning_by_volume.set_for_all_volumes(0)
    water_props.multi_dispense.disposal_by_volume.set_for_all_volumes(disposal_volume)

    # Set distinct air gaps for different post-dispense volumes so that
    # we can verify that air gap volumes are calculated using the current tip volume
    post_aspirate_air_gap_vol = 1.5
    water_props.aspirate.retract.air_gap_by_volume.set_for_all_volumes(
        post_aspirate_air_gap_vol
    )
    water_props.multi_dispense.retract.air_gap_by_volume.set_for_volume(
        400 + disposal_volume, 2.5
    )
    water_props.multi_dispense.retract.air_gap_by_volume.set_for_volume(
        disposal_volume, 3.5
    )
    water_props.multi_dispense.retract.air_gap_by_volume.set_for_volume(0, 4.5)

    with (
        mock.patch.object(
            InstrumentCore,
            "load_liquid_class",
            side_effect=InstrumentCore.load_liquid_class,
            autospec=True,
        ) as patched_load_liquid_class,
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class_during_multi_dispense",
            side_effect=InstrumentCore.dispense_liquid_class_during_multi_dispense,
            autospec=True,
        ) as patched_dispense,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_load_liquid_class, "load_liquid_class")
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(
            patched_dispense, "dispense_liquid_class_during_multi_dispense"
        )
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_1k.distribute_with_liquid_class(
            liquid_class=water,
            volume=400,
            source=nest_plate.rows()[0][1],
            dest=arma_plate.rows()[0][:4],
            new_tip="once",
            trash_location=trash,
        )
        expected_calls = [
            mock.call.load_liquid_class(
                mock.ANY,
                name="water",
                transfer_properties=mock.ANY,
                tiprack_uri="opentrons/opentrons_flex_96_tiprack_1000ul/1",
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=800 + disposal_volume,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                conditioning_volume=0,
                volume_for_pipette_mode_configuration=400,
            ),
            mock.call.dispense_liquid_class_during_multi_dispense(
                mock.ANY,
                volume=400,
                dest=(
                    Location(Point(), arma_plate.rows()[0][0]),
                    arma_plate.rows()[0][0]._core,
                ),
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[
                    LiquidAndAirGapPair(
                        liquid=800 + disposal_volume, air_gap=post_aspirate_air_gap_vol
                    )
                ],
                add_final_air_gap=True,
                trash_location=mock.ANY,
                conditioning_volume=0,
                disposal_volume=disposal_volume,
                is_last_dispense_in_tip=False,
            ),
            mock.call.dispense_liquid_class_during_multi_dispense(
                mock.ANY,
                volume=400,
                dest=(
                    Location(Point(), arma_plate.rows()[0][1]),
                    arma_plate.rows()[0][1]._core,
                ),
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[
                    LiquidAndAirGapPair(liquid=400 + disposal_volume, air_gap=2.5)
                ],
                add_final_air_gap=True,
                trash_location=mock.ANY,
                conditioning_volume=0,
                disposal_volume=disposal_volume,
                is_last_dispense_in_tip=True,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=800 + disposal_volume,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=4.5)],
                conditioning_volume=0,
                volume_for_pipette_mode_configuration=400,
            ),
            mock.call.dispense_liquid_class_during_multi_dispense(
                mock.ANY,
                volume=400,
                dest=(
                    Location(Point(), arma_plate.rows()[0][2]),
                    arma_plate.rows()[0][2]._core,
                ),
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[
                    LiquidAndAirGapPair(
                        liquid=800 + disposal_volume, air_gap=post_aspirate_air_gap_vol
                    )
                ],
                add_final_air_gap=True,
                trash_location=mock.ANY,
                conditioning_volume=0,
                disposal_volume=disposal_volume,
                is_last_dispense_in_tip=False,
            ),
            mock.call.dispense_liquid_class_during_multi_dispense(
                mock.ANY,
                volume=400,
                dest=(
                    Location(Point(), arma_plate.rows()[0][3]),
                    arma_plate.rows()[0][3]._core,
                ),
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[
                    LiquidAndAirGapPair(liquid=400 + disposal_volume, air_gap=2.5)
                ],
                add_final_air_gap=True,
                trash_location=mock.ANY,
                conditioning_volume=0,
                disposal_volume=disposal_volume,
                is_last_dispense_in_tip=True,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
        ]
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
@pytest.mark.parametrize(
    ["distribute_volume", "multi_dispense_props_present"],
    [
        (20, False),
        (26, True),
    ],  # Settings that should result in non-multi-dispense transfers
)
def test_order_of_water_distribute_steps_using_one_to_one_transfers(
    simulated_protocol_context: ProtocolContext,
    distribute_volume: float,
    multi_dispense_props_present: bool,
) -> None:
    """It should distribute liquid using the one-to-one transfer steps instead of doing multi-dispense."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    pipette_50 = simulated_protocol_context.load_instrument(
        "flex_1channel_50", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )
    water = simulated_protocol_context.get_liquid_class("water")
    water_props = water.get_for(pipette_50, tiprack)
    if not multi_dispense_props_present:
        water_props._multi_dispense = None
    expected_post_aspirate_air_gap = (
        water_props.aspirate.retract.air_gap_by_volume.get_for_volume(distribute_volume)
    )
    expected_post_dispense_air_gap = (
        water_props.dispense.retract.air_gap_by_volume.get_for_volume(0)
    )
    with (
        mock.patch.object(
            InstrumentCore,
            "load_liquid_class",
            side_effect=InstrumentCore.load_liquid_class,
            autospec=True,
        ) as patched_load_liquid_class,
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class",
            side_effect=InstrumentCore.dispense_liquid_class,
            autospec=True,
        ) as patched_dispense,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_load_liquid_class, "load_liquid_class")
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(patched_dispense, "dispense_liquid_class")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_50.distribute_with_liquid_class(
            liquid_class=water,
            volume=distribute_volume,
            source=nest_plate.rows()[0][2],
            dest=arma_plate.wells()[:2],
            new_tip="once",
            trash_location=trash,
        )
        expected_calls = [
            mock.call.load_liquid_class(
                mock.ANY,
                name="water",
                transfer_properties=mock.ANY,
                tiprack_uri="opentrons/opentrons_flex_96_tiprack_50ul/1",
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=distribute_volume,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=distribute_volume,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=distribute_volume,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[
                    LiquidAndAirGapPair(
                        liquid=distribute_volume, air_gap=expected_post_aspirate_air_gap
                    )
                ],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=distribute_volume,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[
                    LiquidAndAirGapPair(
                        liquid=0, air_gap=expected_post_dispense_air_gap
                    )
                ],
                volume_for_pipette_mode_configuration=distribute_volume,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=distribute_volume,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[
                    LiquidAndAirGapPair(
                        liquid=distribute_volume, air_gap=expected_post_aspirate_air_gap
                    )
                ],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
        ]
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_order_of_water_distribution_steps_using_mixed_dispense(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should distribute the liquid using multi-dispense and single-dispense steps in the expected order."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_1000ul", "D1"
    )
    pipette_1k = simulated_protocol_context.load_instrument(
        "flex_1channel_1000", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    water_props = water.get_for(pipette_1k, tiprack)
    assert water_props.multi_dispense is not None
    water_props.multi_dispense.retract.blowout.location = "destination"
    water_props.multi_dispense.retract.blowout.flow_rate = pipette_1k.flow_rate.blow_out
    water_props.multi_dispense.retract.blowout.enabled = True
    expected_conditioning_volume = (
        water_props.multi_dispense.conditioning_by_volume.get_for_volume(800)
    )
    expected_disposal_volume = (
        water_props.multi_dispense.disposal_by_volume.get_for_volume(800)
    )
    expected_air_gap = water_props.aspirate.retract.air_gap_by_volume.get_for_volume(
        400
    )
    expected_post_dispense_air_gap = (
        water_props.multi_dispense.retract.air_gap_by_volume.get_for_volume(0)
    )
    with (
        mock.patch.object(
            InstrumentCore,
            "load_liquid_class",
            side_effect=InstrumentCore.load_liquid_class,
            autospec=True,
        ) as patched_load_liquid_class,
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class",
            side_effect=InstrumentCore.dispense_liquid_class,
            autospec=True,
        ) as patched_single_dispense,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class_during_multi_dispense",
            side_effect=InstrumentCore.dispense_liquid_class_during_multi_dispense,
            autospec=True,
        ) as patched_multi_dispense,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_load_liquid_class, "load_liquid_class")
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(patched_single_dispense, "dispense_liquid_class")
        mock_manager.attach_mock(
            patched_multi_dispense, "dispense_liquid_class_during_multi_dispense"
        )
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_1k.distribute_with_liquid_class(
            liquid_class=water,
            volume=400,
            source=nest_plate.rows()[0][1],
            dest=arma_plate.rows()[0][:3],
            new_tip="once",
            trash_location=trash,
        )
        expected_calls = [
            mock.call.load_liquid_class(
                mock.ANY,
                name="water",
                transfer_properties=mock.ANY,
                tiprack_uri="opentrons/opentrons_flex_96_tiprack_1000ul/1",
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=800 + expected_conditioning_volume + expected_disposal_volume,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                conditioning_volume=expected_conditioning_volume,
                volume_for_pipette_mode_configuration=400,
            ),
            mock.call.dispense_liquid_class_during_multi_dispense(
                mock.ANY,
                volume=400,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[
                    LiquidAndAirGapPair(
                        liquid=800 + expected_disposal_volume, air_gap=0
                    )
                ],
                add_final_air_gap=True,
                trash_location=mock.ANY,
                conditioning_volume=expected_conditioning_volume,
                disposal_volume=expected_disposal_volume,
                is_last_dispense_in_tip=False,
            ),
            mock.call.dispense_liquid_class_during_multi_dispense(
                mock.ANY,
                volume=400,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[
                    LiquidAndAirGapPair(
                        liquid=400 + expected_disposal_volume, air_gap=0
                    )
                ],
                add_final_air_gap=True,
                trash_location=mock.ANY,
                conditioning_volume=expected_conditioning_volume,
                disposal_volume=expected_disposal_volume,
                is_last_dispense_in_tip=True,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=400,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[
                    LiquidAndAirGapPair(
                        liquid=0, air_gap=expected_post_dispense_air_gap
                    )
                ],
                conditioning_volume=0,
                volume_for_pipette_mode_configuration=400,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=400,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_MANY,
                tip_contents=[
                    LiquidAndAirGapPair(
                        liquid=400,
                        air_gap=expected_air_gap,
                    )
                ],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
        ]
        assert mock_manager.mock_calls[1] == expected_calls[1]
        assert mock_manager.mock_calls[2] == expected_calls[2]
        assert mock_manager.mock_calls[3] == expected_calls[3]
        assert mock_manager.mock_calls[4] == expected_calls[4]
        assert mock_manager.mock_calls[5] == expected_calls[5]
        assert mock_manager.mock_calls[6] == expected_calls[6]
        assert mock_manager.mock_calls[7] == expected_calls[7]
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_water_distribute_steps_with_return_tip(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should return tips during the distribution.

    This test uses the same liquid class and transfer parameters as
    `test_order_of_water_distribution_steps_using_mixed_dispense`, except that it sets
    `return_tip` arg to True. So we expect the execution to call `drop_tip()` with
    the well_core pointing to last tip's well and a `location` of None.
    """
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_1000ul", "D1"
    )
    pipette_1k = simulated_protocol_context.load_instrument(
        "flex_1channel_1000", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )
    water = simulated_protocol_context.get_liquid_class("water")

    with (
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "drop_tip",
            side_effect=InstrumentCore.drop_tip,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip")
        pipette_1k.distribute_with_liquid_class(
            liquid_class=water,
            volume=400,
            source=nest_plate.rows()[0][1],
            dest=arma_plate.rows()[0][:3],
            new_tip="once",
            trash_location=trash,
            return_tip=True,
        )
        expected_calls = [
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.drop_tip(
                mock.ANY,
                location=None,
                well_core=mock.ANY,
                home_after=False,
                alternate_drop_location=False,
            ),
        ]
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_water_distribution_raises_error_for_disposal_vol_without_blowout(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should execute the distribute steps with the expected tip pick ups."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_1000ul", "D1"
    )
    pipette_1k = simulated_protocol_context.load_instrument(
        "flex_1channel_1000", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    water = simulated_protocol_context.get_liquid_class("water")
    water_props = water.get_for(pipette_1k, tiprack)
    water_props.multi_dispense.retract.blowout.enabled = False  # type: ignore[union-attr]
    with pytest.raises(
        RuntimeError,
        match="Specify a blowout location and enable blowout when using a disposal volume",
    ):
        pipette_1k.distribute_with_liquid_class(
            liquid_class=water,
            volume=140,
            source=nest_plate.rows()[0][0],
            dest=nest_plate.rows()[1],
            new_tip="once",
            trash_location=trash,
        )


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_transfers_do_not_perform_lpd(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should not do any liquid probing for LC-based transfer/ consolidate/ distribute."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_1000ul", "D1"
    )
    pipette_1k = simulated_protocol_context.load_instrument(
        "flex_1channel_1000",
        mount="left",
        tip_racks=[tiprack],
        liquid_presence_detection=True,
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )
    water = simulated_protocol_context.get_liquid_class("water")

    with mock.patch.object(
        InstrumentCore,
        "liquid_probe_with_recovery",
        autospec=True,
    ) as patched_liquid_probe:
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_liquid_probe, "liquid_probe_with_recovery")
        pipette_1k.transfer_with_liquid_class(
            liquid_class=water,
            volume=1100,
            source=nest_plate.rows()[0],
            dest=arma_plate.rows()[0],
            new_tip="always",
            trash_location=trash,
        )
        pipette_1k.consolidate_with_liquid_class(
            liquid_class=water,
            volume=100,
            source=nest_plate.rows()[0],
            dest=arma_plate.rows()[0][0],
            new_tip="always",
            trash_location=trash,
        )
        pipette_1k.distribute_with_liquid_class(
            liquid_class=water,
            volume=100,
            source=nest_plate.rows()[0][0],
            dest=arma_plate.rows()[0],
            new_tip="always",
            trash_location=trash,
        )
        assert patched_liquid_probe.call_count == 0


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_water_transfer_with_multi_channel_pipette(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should run the transfer steps for a multi-channel pipette and well grouping without any errors.

    This test only checks that various supported configurations for a transfer
    analyze successfully. It doesn't check whether the steps are as expected.
    That will be covered in analysis snapshot tests.
    """
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    pipette_50 = simulated_protocol_context.load_instrument(
        "flex_8channel_50", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    with (
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class",
            side_effect=InstrumentCore.dispense_liquid_class,
            autospec=True,
        ) as patched_dispense,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(patched_dispense, "dispense_liquid_class")
        pipette_50.transfer_with_liquid_class(
            liquid_class=water,
            volume=40,
            source=nest_plate.columns()[:2],
            dest=arma_plate.columns()[:2],
            new_tip="always",
            trash_location=trash,
        )
        assert patched_aspirate.call_count == 2
        assert patched_dispense.call_count == 2


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_raises_no_tips_available_error(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should raise an error explaining that there aren't any tips available."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack1 = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    tiprack2 = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D2"
    )
    pipette_50 = simulated_protocol_context.load_instrument(
        "flex_1channel_50", mount="left", tip_racks=[tiprack1, tiprack2]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )
    water = simulated_protocol_context.get_liquid_class("water")
    expected_error_msg = (
        "No tip available among the tipracks assigned for flex_1channel_50:"
        " \\['Opentrons Flex 96 Tip Rack 50 µL in D1', 'Opentrons Flex 96 Tip Rack 50 µL in D2'\\]"
    )
    with pytest.raises(RuntimeError, match=expected_error_msg):
        pipette_50.transfer_with_liquid_class(
            liquid_class=water,
            volume=160,
            source=nest_plate.columns(),
            dest=arma_plate.columns(),
            new_tip="always",
            trash_location=trash,
        )
    with pytest.raises(RuntimeError, match=f"{expected_error_msg}"):
        pipette_50.distribute_with_liquid_class(
            liquid_class=water,
            volume=160,
            source=nest_plate.wells()[-1],
            dest=arma_plate.columns(),
            new_tip="once",
            trash_location=trash,
        )
    with pytest.raises(RuntimeError, match=f"{expected_error_msg}"):
        pipette_50.consolidate_with_liquid_class(
            liquid_class=water,
            volume=50,
            source=nest_plate.columns(),
            dest=arma_plate.wells()[0],
            new_tip="once",
            trash_location=trash,
        )


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_transfer_with_keep_last_tip(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should keep the last tip instead of dropping the last one."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_1000ul", "D1"
    )
    pipette_1k = simulated_protocol_context.load_instrument(
        "flex_1channel_1000", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )
    water = simulated_protocol_context.get_liquid_class("water")

    with (
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_1k.transfer_with_liquid_class(
            liquid_class=water,
            volume=400,
            source=nest_plate.wells()[:2],
            dest=arma_plate.wells()[:2],
            new_tip="always",
            trash_location=trash,
            keep_last_tip=True,
        )
        expected_calls = [
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
        ]
        assert pipette_1k.has_tip
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_transfer_with_keep_last_tip_false(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should not keep the last tip for a tip policy of never."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_1000ul", "D1"
    )
    pipette_1k = simulated_protocol_context.load_instrument(
        "flex_1channel_1000", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )
    water = simulated_protocol_context.get_liquid_class("water")

    with (
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        pipette_1k.pick_up_tip()
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_1k.transfer_with_liquid_class(
            liquid_class=water,
            volume=400,
            source=nest_plate.wells()[:2],
            dest=arma_plate.wells()[:2],
            new_tip="never",
            trash_location=trash,
            keep_last_tip=False,
        )
        expected_calls = [
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            )
        ]
        assert not pipette_1k.has_tip
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_transfer_with_keep_last_tip_chained(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should be able to use the last kept tip when calling with never after another transfer."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_1000ul", "D1"
    )
    pipette_1k = simulated_protocol_context.load_instrument(
        "flex_1channel_1000", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )
    water = simulated_protocol_context.get_liquid_class("water")

    with (
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_1k.distribute_with_liquid_class(
            liquid_class=water,
            volume=400,
            source=nest_plate.wells()[0],
            dest=arma_plate.wells()[:2],
            new_tip="always",
            trash_location=trash,
            keep_last_tip=True,
        )
        assert pipette_1k.has_tip
        pipette_1k.consolidate_with_liquid_class(
            liquid_class=water,
            volume=400,
            source=nest_plate.wells()[:2],
            dest=arma_plate.wells()[0],
            new_tip="never",
            trash_location=trash,
        )
        expected_calls = [
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            )
        ]
        assert pipette_1k.has_tip
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_return_tip_after_transfer_with_never(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should be able to return the current tip when calling a transfer with return tip."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_1000ul", "D1"
    )
    pipette_1k = simulated_protocol_context.load_instrument(
        "flex_1channel_1000", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )
    water = simulated_protocol_context.get_liquid_class("water")

    with (
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "drop_tip",
            side_effect=InstrumentCore.drop_tip,
            autospec=True,
        ) as patched_drop_tip,
    ):
        pipette_1k.pick_up_tip()
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip")
        pipette_1k.transfer_with_liquid_class(
            liquid_class=water,
            volume=400,
            source=nest_plate.wells()[:2],
            dest=arma_plate.wells()[:2],
            new_tip="never",
            trash_location=trash,
            keep_last_tip=False,
            return_tip=True,
        )
        expected_calls = [
            mock.call.drop_tip(
                mock.ANY,
                location=None,
                well_core=mock.ANY,
                home_after=False,
                alternate_drop_location=False,
            ),
        ]
        assert not pipette_1k.has_tip
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_return_tip_after_chained_transfers(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should be able to return the current tip when calling a transfer with return tip."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_1000ul", "D1"
    )
    pipette_1k = simulated_protocol_context.load_instrument(
        "flex_1channel_1000", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )
    water = simulated_protocol_context.get_liquid_class("water")

    with (
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "drop_tip",
            side_effect=InstrumentCore.drop_tip,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip")
        pipette_1k.transfer_with_liquid_class(
            liquid_class=water,
            volume=400,
            source=nest_plate.wells()[:2],
            dest=arma_plate.wells()[:2],
            new_tip="once",
            trash_location=trash,
            keep_last_tip=True,
        )
        assert pipette_1k.has_tip
        pipette_1k.distribute_with_liquid_class(
            liquid_class=water,
            volume=400,
            source=nest_plate.wells()[0],
            dest=arma_plate.wells()[:2],
            new_tip="never",
            trash_location=trash,
            keep_last_tip=False,
            return_tip=True,
        )
        expected_calls = [
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.drop_tip(
                mock.ANY,
                location=None,
                well_core=mock.ANY,
                home_after=False,
                alternate_drop_location=False,
            ),
        ]
        assert not pipette_1k.has_tip
        assert mock_manager.mock_calls == expected_calls


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_return_tip_fails_after_new_tip_never_keep_last_tip_false(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should fail in returning a tip with new tip of never and keep last tip False."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_1000ul", "D1"
    )
    pipette_1k = simulated_protocol_context.load_instrument(
        "flex_1channel_1000", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )
    water = simulated_protocol_context.get_liquid_class("water")

    pipette_1k.pick_up_tip()
    pipette_1k.transfer_with_liquid_class(
        liquid_class=water,
        volume=400,
        source=nest_plate.wells()[0],
        dest=arma_plate.wells()[0],
        new_tip="never",
        trash_location=trash,
        keep_last_tip=False,
    )

    with pytest.raises(TypeError, match="Last tip location"):
        pipette_1k.return_tip()


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.27", "Flex")], indirect=True
)
def test_water_transfer_with_selected_tips(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should run the transfer using the tips provided."""
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    pipette_50 = simulated_protocol_context.load_instrument(
        "flex_1channel_50",
        mount="left",
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    with (
        mock.patch.object(
            InstrumentCore,
            "load_liquid_class",
            side_effect=InstrumentCore.load_liquid_class,
            autospec=True,
        ) as patched_load_liquid_class,
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class",
            side_effect=InstrumentCore.dispense_liquid_class,
            autospec=True,
        ) as patched_dispense,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_load_liquid_class, "load_liquid_class")
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(patched_dispense, "dispense_liquid_class")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_50.transfer_with_liquid_class(
            liquid_class=water,
            volume=40,
            source=nest_plate.rows()[0][:2],
            dest=arma_plate.rows()[0][:2],
            new_tip="always",
            trash_location=trash,
            tips=[tiprack["D10"], tiprack["B3"]],
        )
        expected_calls = [
            mock.call.load_liquid_class(
                mock.ANY,
                name="water",
                transfer_properties=mock.ANY,
                tiprack_uri="opentrons/opentrons_flex_96_tiprack_50ul/1",
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=Location(tiprack["D10"]._core.get_top(0), tiprack),
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=40,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=40,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=40,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=40, air_gap=0.1)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=Location(tiprack["B3"]._core.get_top(0), tiprack),
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=40,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=40,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=40,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=40, air_gap=0.1)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
        ]
        assert mock_manager.mock_calls == expected_calls


# Fix this test
@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.24", "Flex")], indirect=True
)
def test_order_of_water_transfer_steps_with_blowout_in_source_custom_position(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should run the transfer steps with blowout at specified position in source labware.

    This test only checks that various supported configurations for a transfer
    analyze successfully. It doesn't check whether the steps are as expected.
    That will be covered in analysis snapshot tests.
    """
    trash = simulated_protocol_context.load_trash_bin("A3")
    tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D1"
    )
    pipette_50 = simulated_protocol_context.load_instrument(
        "flex_1channel_50", mount="left", tip_racks=[tiprack]
    )
    nest_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "C3"
    )
    arma_plate = simulated_protocol_context.load_labware(
        "armadillo_96_wellplate_200ul_pcr_full_skirt", "C2"
    )

    water = simulated_protocol_context.get_liquid_class("water")
    water_blowout_props = water.get_for(pipette_50, tiprack).dispense.retract.blowout
    water_blowout_props.enabled = True
    water_blowout_props.location = "source"
    water_blowout_props.blowout_position = {
        "position_reference": "well-top",
        "offset": {"x": 1, "y": 2, "z": 3},
    }

    with (
        mock.patch.object(
            InstrumentCore,
            "load_liquid_class",
            side_effect=InstrumentCore.load_liquid_class,
            autospec=True,
        ) as patched_load_liquid_class,
        mock.patch.object(
            InstrumentCore,
            "pick_up_tip",
            side_effect=InstrumentCore.pick_up_tip,
            autospec=True,
        ) as patched_pick_up_tip,
        mock.patch.object(
            InstrumentCore,
            "aspirate_liquid_class",
            side_effect=InstrumentCore.aspirate_liquid_class,
            autospec=True,
        ) as patched_aspirate,
        mock.patch.object(
            InstrumentCore,
            "dispense_liquid_class",
            side_effect=InstrumentCore.dispense_liquid_class,
            autospec=True,
        ) as patched_dispense,
        mock.patch.object(
            InstrumentCore,
            "blow_out",
            side_effect=InstrumentCore.blow_out,
            autospec=True,
        ) as patched_blowout,
        mock.patch.object(
            InstrumentCore,
            "drop_tip_in_disposal_location",
            side_effect=InstrumentCore.drop_tip_in_disposal_location,
            autospec=True,
        ) as patched_drop_tip,
    ):
        mock_manager = mock.Mock()
        mock_manager.attach_mock(patched_pick_up_tip, "pick_up_tip")
        mock_manager.attach_mock(patched_load_liquid_class, "load_liquid_class")
        mock_manager.attach_mock(patched_aspirate, "aspirate_liquid_class")
        mock_manager.attach_mock(patched_dispense, "dispense_liquid_class")
        mock_manager.attach_mock(patched_blowout, "blow_out")
        mock_manager.attach_mock(patched_drop_tip, "drop_tip_in_disposal_location")
        pipette_50.transfer_with_liquid_class(
            liquid_class=water,
            volume=40,
            source=nest_plate.rows()[0][:2],
            dest=arma_plate.rows()[0][:2],
            new_tip="always",
            trash_location=trash,
        )
        expected_calls = [
            mock.call.load_liquid_class(
                mock.ANY,
                name="water",
                transfer_properties=mock.ANY,
                tiprack_uri="opentrons/opentrons_flex_96_tiprack_50ul/1",
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=40,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=40,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=40,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=40, air_gap=0.1)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.blow_out(
                mock.ANY,
                location=Location(
                    nest_plate.rows()[0][0].top(0).point + Point(1, 2, 3),
                    labware=nest_plate.rows()[0][0],
                ),
                well_core=mock.ANY,
                in_place=False,
                flow_rate=50,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
            mock.call.pick_up_tip(
                mock.ANY,
                location=mock.ANY,
                well_core=mock.ANY,
                presses=mock.ANY,
                increment=mock.ANY,
            ),
            mock.call.aspirate_liquid_class(
                mock.ANY,
                volume=40,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=0, air_gap=0)],
                volume_for_pipette_mode_configuration=40,
            ),
            mock.call.dispense_liquid_class(
                mock.ANY,
                volume=40,
                dest=mock.ANY,
                source=mock.ANY,
                transfer_properties=mock.ANY,
                transfer_type=TransferType.ONE_TO_ONE,
                tip_contents=[LiquidAndAirGapPair(liquid=40, air_gap=0.1)],
                add_final_air_gap=True,
                trash_location=mock.ANY,
            ),
            mock.call.blow_out(
                mock.ANY,
                location=Location(
                    nest_plate.rows()[0][1].top(0).point + Point(1, 2, 3),
                    labware=nest_plate.rows()[0][1],
                ),
                well_core=mock.ANY,
                in_place=False,
                flow_rate=50,
            ),
            mock.call.drop_tip_in_disposal_location(
                mock.ANY,
                disposal_location=trash,
                home_after=False,
                alternate_tip_drop=True,
            ),
        ]
        assert mock_manager.mock_calls == expected_calls



import asyncio
import threading
import weakref
from typing import NoReturn, Optional

import pytest

from opentrons.hardware_control.api import API
from opentrons.hardware_control.modules import ModuleAtPort, SimulatingModule
from opentrons.hardware_control.thread_manager import (
    ThreadManager,
    ThreadManagerException,
)


async def _raise_exception() -> NoReturn:
    raise Exception("oh no")


def test_build_fail_raises_exception() -> None:
    """
    Test that a builder that raises an exception raises
    a ThreadManagerException
    """
    with pytest.raises(ThreadManagerException):
        ThreadManager(_raise_exception)


def test_module_cache_add_entry() -> None:
    """Test that _cached_modules updates correctly."""

    mod_names = {
        "tempdeck": [SimulatingModule(serial_number="111", model="temperatureModuleV2")]
    }
    thread_manager = ThreadManager(
        API.build_hardware_simulator, attached_modules=mod_names
    )

    # Test that module gets added to the cache
    mods = thread_manager.attached_modules
    wrapper_cache = thread_manager._cached_modules.copy()
    assert isinstance(wrapper_cache, weakref.WeakKeyDictionary)
    assert len(wrapper_cache) == 1
    assert mods[0] in wrapper_cache.values()

    # Test that calling attached modules doesn't add duplicate entries to cache
    mods2 = thread_manager.attached_modules
    wrapper_cache2 = thread_manager._cached_modules.copy()
    assert len(wrapper_cache2) == 1
    assert mods == mods2


async def test_module_cache_remove_entry() -> None:
    """Test that module entry gets removed from cache when module detaches."""
    mod_names = {
        "tempdeck": [
            SimulatingModule(serial_number="111", model="temperatureModuleV2")
        ],
        "magdeck": [SimulatingModule(serial_number="222", model="magneticModuleV1")],
    }
    thread_manager = ThreadManager(
        API.build_hardware_simulator, attached_modules=mod_names
    )

    mods_before = thread_manager.attached_modules
    assert len(mods_before) == 2

    loop = thread_manager._loop
    assert loop

    # The coroutine must be called using the threadmanager's loop.
    future = asyncio.run_coroutine_threadsafe(
        thread_manager._backend.module_controls.register_modules(
            removed_mods_at_ports=[
                ModuleAtPort(port="/dev/ot_module_sim_tempdeck111", name="tempdeck")
            ]
        ),
        loop,
    )
    future.result()
    mods_after = thread_manager.attached_modules
    assert len(mods_after) == 1


async def test_wraps_instance() -> None:
    """It should expose the underlying type."""
    thread_manager = ThreadManager(API.build_hardware_simulator)
    assert thread_manager.wraps_instance(API)


class Blocker:
    """Test object for nonblocking construction."""

    @classmethod
    async def build_blocking(
        cls,
        wait_event: asyncio.Event,
        running_event: threading.Event,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> "Blocker":
        """Build an instance."""
        inst = cls(loop)
        running_event.set()
        await wait_event.wait()
        return inst

    def __init__(self, loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        """Initialize an instance."""
        self._loop = loop

    async def clean_up(self) -> None:
        """Allows cleanup."""
        pass


def test_nonblocking_via_attr() -> None:
    """Its init should return immediately if the builder has a nonblocking attribute."""

    wait_event = asyncio.Event()
    running_event = threading.Event()
    thread_manager = ThreadManager(  # type: ignore[type-var]
        ThreadManager.nonblocking_builder(Blocker.build_blocking),
        running_event=running_event,
        wait_event=wait_event,
    )

    running_event.wait()
    assert not thread_manager._is_running.is_set()
    assert thread_manager._loop
    thread_manager._loop.call_soon_threadsafe(wait_event.set)
    thread_manager.managed_thread_ready_blocking()
    thread_manager.clean_up()



import copy

from .ot3_settings import ot3_dummy_settings
from opentrons.config import defaults_ot3, robot_configs
from opentrons.config.types import GantryLoad, OT3Config
from opentrons.hardware_control.types import OT3AxisKind


def test_load_calibration_cals() -> None:
    # nothing provided
    assert (
        defaults_ot3._build_default_calibration(
            {}, defaults_ot3.DEFAULT_CALIBRATION_SETTINGS
        )
        == defaults_ot3.DEFAULT_CALIBRATION_SETTINGS
    )

    # some dicts not formatted right
    mostly_right = defaults_ot3.serialize(defaults_ot3.build_with_defaults({}))
    del mostly_right["calibration"]["edge_sense"]["early_sense_tolerance_mm"]
    del mostly_right["calibration"]["z_offset"]["pass_settings"]["prep_distance_mm"]
    assert (
        defaults_ot3._build_default_calibration(
            mostly_right["calibration"], defaults_ot3.DEFAULT_CALIBRATION_SETTINGS
        )
        == defaults_ot3.DEFAULT_CALIBRATION_SETTINGS
    )

    # altered values are preserved
    mostly_right["calibration"]["edge_sense"]["overrun_tolerance_mm"] -= 0.2
    mostly_right["calibration"]["z_offset"]["pass_settings"][
        "max_overrun_distance_mm"
    ] -= 0.5
    built_with_overrides = defaults_ot3._build_default_calibration(
        mostly_right["calibration"], defaults_ot3.DEFAULT_CALIBRATION_SETTINGS
    )
    assert (
        built_with_overrides.edge_sense.overrun_tolerance_mm
        == mostly_right["calibration"]["edge_sense"]["overrun_tolerance_mm"]
    )
    assert (
        built_with_overrides.z_offset.pass_settings.max_overrun_distance_mm
        == mostly_right["calibration"]["z_offset"]["pass_settings"][
            "max_overrun_distance_mm"
        ]
    )


def test_load_per_pipette_vals() -> None:
    # nothing provided
    assert (
        defaults_ot3._build_default_bpk({}, defaults_ot3.DEFAULT_HOLD_CURRENT)
        == defaults_ot3.DEFAULT_HOLD_CURRENT
    )

    # some dicts not formatted right
    mostly_right = defaults_ot3.serialize(defaults_ot3.build_with_defaults({}))
    del mostly_right["motion_settings"]["acceleration"]["low_throughput"]
    assert (
        defaults_ot3._build_default_bpk(
            mostly_right["motion_settings"]["acceleration"],
            defaults_ot3.DEFAULT_ACCELERATIONS,
        ).low_throughput
        == defaults_ot3.DEFAULT_ACCELERATIONS.low_throughput
    )

    # altered values are preserved
    mostly_right["motion_settings"]["acceleration"]["high_throughput_1000"]["X"] -= 2
    assert (
        defaults_ot3._build_default_bpk(
            mostly_right["motion_settings"]["acceleration"],
            defaults_ot3.DEFAULT_ACCELERATIONS,
        ).high_throughput_1000[OT3AxisKind.X]
        == defaults_ot3.DEFAULT_ACCELERATIONS.high_throughput_1000[OT3AxisKind.X] - 2
    )

    # added values are preserved
    altered_default = copy.deepcopy(defaults_ot3.DEFAULT_ACCELERATIONS)
    altered_default.high_throughput_1000.pop(OT3AxisKind.X, None)

    mostly_right["motion_settings"]["acceleration"]["high_throughput_1000"]["X"] = -72
    assert (
        defaults_ot3._build_default_bpk(
            mostly_right["motion_settings"]["acceleration"], altered_default
        ).high_throughput_1000[OT3AxisKind.X]
        == -72
    )


def test_load_offset_vals() -> None:
    # nothing provided
    assert (
        defaults_ot3._build_default_offset([], defaults_ot3.DEFAULT_RIGHT_MOUNT_OFFSET)
        == defaults_ot3.DEFAULT_RIGHT_MOUNT_OFFSET
    )
    # wrong shape
    assert (
        defaults_ot3._build_default_offset(
            [
                1,
            ],
            defaults_ot3.DEFAULT_RIGHT_MOUNT_OFFSET,
        )
        == defaults_ot3.DEFAULT_RIGHT_MOUNT_OFFSET
    )
    # absolutely wrong type
    assert (
        defaults_ot3._build_default_offset(2, defaults_ot3.DEFAULT_RIGHT_MOUNT_OFFSET)
        == defaults_ot3.DEFAULT_RIGHT_MOUNT_OFFSET
    )
    # right shape, different values
    assert defaults_ot3._build_default_offset([1, 2, 3], (0, 0, 0)) == (1, 2, 3)


def test_load_transform_vals() -> None:
    # not list
    assert (
        defaults_ot3._build_default_transform(
            None, defaults_ot3.DEFAULT_MACHINE_TRANSFORM
        )
        == defaults_ot3.DEFAULT_MACHINE_TRANSFORM
    )
    # list of not lists
    assert (
        defaults_ot3._build_default_transform(
            [1, 2, 3], defaults_ot3.DEFAULT_MACHINE_TRANSFORM
        )
        == defaults_ot3.DEFAULT_MACHINE_TRANSFORM
    )
    # list of wrong number of lists
    assert (
        defaults_ot3._build_default_transform(
            [[1, 2, 3], [3, 2, 1]], defaults_ot3.DEFAULT_MACHINE_TRANSFORM
        )
        == defaults_ot3.DEFAULT_MACHINE_TRANSFORM
    )
    # list of right number of lists of wrong number of elements
    assert (
        defaults_ot3._build_default_transform(
            [[1, 2, 3], [1, 2, 3], [1, 2, 3, 4]], defaults_ot3.DEFAULT_MACHINE_TRANSFORM
        )
        == defaults_ot3.DEFAULT_MACHINE_TRANSFORM
    )
    # list of right number of lists of right number of elements of wrong type
    assert (
        defaults_ot3._build_default_transform(
            [
                [
                    1,
                    2,
                    3,
                ],
                [1, "hi", 3],
                [1, 2, {}],
            ],
            defaults_ot3.DEFAULT_MACHINE_TRANSFORM,
        )
        == defaults_ot3.DEFAULT_MACHINE_TRANSFORM
    )
    # right shape, different data is preserved
    assert defaults_ot3._build_default_transform(
        [[1, 2, 3], [1, 2, 3], [1, 2, 3]], defaults_ot3.DEFAULT_MACHINE_TRANSFORM
    ) == [[1, 2, 3], [1, 2, 3], [1, 2, 3]]


def test_motion_settings_dataclass() -> None:
    built_config = robot_configs.build_config(ot3_dummy_settings)
    assert isinstance(built_config, OT3Config)
    motion_settings = built_config.motion_settings

    low_setting = motion_settings.by_gantry_load(GantryLoad.LOW_THROUGHPUT)
    assert low_setting["acceleration"] == {
        OT3AxisKind.X: 3,
        OT3AxisKind.Y: 2,
        OT3AxisKind.Z: 15,
        OT3AxisKind.P: 15,
        OT3AxisKind.Z_G: 5,
    }
    assert low_setting["default_max_speed"] == {
        OT3AxisKind.X: 1,
        OT3AxisKind.Y: 2,
        OT3AxisKind.Z: 3,
        OT3AxisKind.P: 4,
        OT3AxisKind.Z_G: 5,
    }
    assert low_setting["max_speed_discontinuity"] == {
        OT3AxisKind.X: 10,
        OT3AxisKind.Y: 20,
        OT3AxisKind.Z: 30,
        OT3AxisKind.P: 40,
        OT3AxisKind.Z_G: 50,
    }
    assert low_setting["direction_change_speed_discontinuity"] == {
        OT3AxisKind.X: 5,
        OT3AxisKind.Y: 10,
        OT3AxisKind.Z: 15,
        OT3AxisKind.P: 20,
        OT3AxisKind.Z_G: 15,
    }



"""Tests for Protocol API module contexts."""

from typing import cast

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.types import LabwareDefinition as LabwareDefDict

from opentrons.hardware_control.modules.types import HeaterShakerModuleModel, ModuleType
from opentrons.legacy_broker import LegacyBroker
from opentrons.protocol_api import MAX_SUPPORTED_VERSION, Labware, ModuleContext
from opentrons.protocol_api.core.common import LabwareCore, ModuleCore, ProtocolCore
from opentrons.protocol_api.core.core_map import LoadedCoreMap
from opentrons.protocol_api.core.labware import LabwareLoadParams
from opentrons.protocols.api_support.types import APIVersion


@pytest.fixture
def mock_core(decoy: Decoy) -> ModuleCore:
    """Get a mock module implementation core."""
    return decoy.mock(cls=ModuleCore)


@pytest.fixture
def mock_core_map(decoy: Decoy) -> LoadedCoreMap:
    """Get a mock core map."""
    return decoy.mock(cls=LoadedCoreMap)


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock protocol implementation core."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture
def mock_broker(decoy: Decoy) -> LegacyBroker:
    """Get a mock command message broker."""
    return decoy.mock(cls=LegacyBroker)


@pytest.fixture
def api_version() -> APIVersion:
    """Set the API version for the test."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def subject(
    mock_core: ModuleCore,
    mock_core_map: LoadedCoreMap,
    mock_protocol_core: ProtocolCore,
    mock_broker: LegacyBroker,
    api_version: APIVersion,
) -> ModuleContext:
    """Get a generic module context with its dependencies mocked out."""
    return ModuleContext(
        core=mock_core,
        core_map=mock_core_map,
        protocol_core=mock_protocol_core,
        broker=mock_broker,
        api_version=api_version,
    )


def test_get_labware(
    decoy: Decoy,
    mock_core: ModuleCore,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    subject: ModuleContext,
) -> None:
    """It should return the labware from the protocol core's loaded equipment object."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    mock_labware = decoy.mock(cls=Labware)

    decoy.when(mock_protocol_core.get_labware_on_module(mock_core)).then_return(
        mock_labware_core
    )
    decoy.when(mock_core_map.get(mock_labware_core)).then_return(mock_labware)

    assert subject.labware is mock_labware


@pytest.mark.parametrize("api_version", [APIVersion(2, 1234)])
def test_load_labware(
    decoy: Decoy,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    mock_core: ModuleCore,
    api_version: APIVersion,
    subject: ModuleContext,
) -> None:
    """It should load labware by load parameters."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(
        mock_protocol_core.load_labware(
            load_name="infinite tip rack",
            label="it doesn't run out",
            namespace="ideal",
            version=101,
            location=mock_core,
        )
    ).then_return(mock_labware_core)

    decoy.when(mock_labware_core.get_name()).then_return("Full Name")
    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    result = subject.load_labware(
        name="Infinite Tip Rack",
        label="it doesn't run out",
        namespace="ideal",
        version=101,
    )

    assert isinstance(result, Labware)
    assert result.name == "Full Name"
    assert result.api_version == api_version
    decoy.verify(mock_core_map.add(mock_labware_core, result), times=1)


@pytest.mark.parametrize("api_version", [APIVersion(2, 1234)])
@pytest.mark.parametrize(
    "label,sanitized_label", [(7, "7"), ("hi", "hi"), (None, None)]
)
def test_load_labware_sanitizes_label(
    decoy: Decoy,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    mock_core: ModuleCore,
    api_version: APIVersion,
    label: str | None,
    sanitized_label: str | None,
    subject: ModuleContext,
) -> None:
    """It should load labware by load parameters."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(
        mock_protocol_core.load_labware(
            load_name="infinite tip rack",
            label=sanitized_label,
            namespace="ideal",
            version=101,
            location=mock_core,
        )
    ).then_return(mock_labware_core)

    decoy.when(mock_labware_core.get_name()).then_return("Full Name")
    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    result = subject.load_labware(
        name="Infinite Tip Rack",
        label=label,
        namespace="ideal",
        version=101,
    )

    assert isinstance(result, Labware)
    assert result.name == "Full Name"
    assert result.api_version == api_version
    decoy.verify(mock_core_map.add(mock_labware_core, result), times=1)


@pytest.mark.parametrize("api_version", [APIVersion(2, 1234)])
def test_load_labware_from_definition(
    decoy: Decoy,
    mock_core: ModuleCore,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ModuleContext,
) -> None:
    """It should be able to load a labware from a definition dictionary."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    labware_definition_dict = cast(LabwareDefDict, {"labwareDef": True})
    labware_load_params = LabwareLoadParams("you", "are", 1337)

    decoy.when(
        mock_protocol_core.add_labware_definition(labware_definition_dict)
    ).then_return(labware_load_params)

    decoy.when(mock_labware_core.get_name()).then_return("Full Name")
    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    decoy.when(
        mock_protocol_core.load_labware(
            namespace="you",
            load_name="are",
            version=1337,
            label="Some Display Name",
            location=mock_core,
        )
    ).then_return(mock_labware_core)

    result = subject.load_labware_from_definition(
        definition=labware_definition_dict,
        label="Some Display Name",
    )

    assert isinstance(result, Labware)
    assert result.name == "Full Name"
    assert result.api_version == api_version
    decoy.verify(mock_core_map.add(mock_labware_core, result), times=1)


def test_load_adapter(
    decoy: Decoy,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    mock_core: ModuleCore,
    subject: ModuleContext,
) -> None:
    """It should load adapter by load parameters."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(
        mock_protocol_core.load_adapter(
            load_name="an adapter",
            namespace="spacename",
            version=101,
            location=mock_core,
        )
    ).then_return(mock_labware_core)

    decoy.when(mock_labware_core.get_name()).then_return("Full Name")
    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    result = subject.load_adapter(
        name="an adapter",
        namespace="spacename",
        version=101,
    )

    assert isinstance(result, Labware)
    assert result.name == "Full Name"
    decoy.verify(mock_core_map.add(mock_labware_core, result), times=1)


def test_load_adapter_from_definition(
    decoy: Decoy,
    mock_core: ModuleCore,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    subject: ModuleContext,
) -> None:
    """It should be able to load a labware from a definition dictionary."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    labware_definition_dict = cast(LabwareDefDict, {"labwareDef": True})
    labware_load_params = LabwareLoadParams("you", "are", 1337)

    decoy.when(
        mock_protocol_core.add_labware_definition(labware_definition_dict)
    ).then_return(labware_load_params)

    decoy.when(mock_labware_core.get_name()).then_return("Full Name")
    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    decoy.when(
        mock_protocol_core.load_adapter(
            namespace="you",
            load_name="are",
            version=1337,
            location=mock_core,
        )
    ).then_return(mock_labware_core)

    result = subject.load_adapter_from_definition(
        definition=labware_definition_dict,
    )

    assert isinstance(result, Labware)
    assert result.name == "Full Name"
    decoy.verify(mock_core_map.add(mock_labware_core, result), times=1)


@pytest.mark.parametrize("api_version", [APIVersion(2, 1234)])
def test_load_labware_with_adapter(
    decoy: Decoy,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    mock_core: ModuleCore,
    api_version: APIVersion,
    subject: ModuleContext,
) -> None:
    """It should load labware by load parameters."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    mock_adapter_core = decoy.mock(cls=LabwareCore)

    decoy.when(
        mock_protocol_core.load_adapter(
            load_name="adaptation",
            namespace=None,
            version=None,
            location=mock_core,
        )
    ).then_return(mock_adapter_core)

    decoy.when(mock_adapter_core.get_well_columns()).then_return([])

    decoy.when(
        mock_protocol_core.load_labware(
            load_name="the platonic labware",
            label="perfect",
            namespace="ideal",
            version=101,
            location=mock_adapter_core,
        )
    ).then_return(mock_labware_core)

    decoy.when(mock_labware_core.get_name()).then_return("Full Name")
    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    result = subject.load_labware(
        name="the platonic labware",
        label="perfect",
        namespace="ideal",
        version=101,
        adapter="adaptation",
    )

    assert isinstance(result, Labware)
    assert result.name == "Full Name"
    assert result.api_version == api_version
    decoy.verify(mock_core_map.add(mock_labware_core, result), times=1)


def test_parent(decoy: Decoy, mock_core: ModuleCore, subject: ModuleContext) -> None:
    """Should get the parent slot name."""
    decoy.when(mock_core.get_deck_slot_id()).then_return("bar")
    assert subject.parent == "bar"


def test_module_model(
    decoy: Decoy,
    mock_core: ModuleCore,
    subject: ModuleContext,
) -> None:
    """It should get module's model."""
    decoy.when(mock_core.get_model()).then_return(
        HeaterShakerModuleModel("heaterShakerModuleV1")
    )
    result = subject.model
    assert result == "heaterShakerModuleV1"


def test_module_type(
    decoy: Decoy,
    mock_core: ModuleCore,
    subject: ModuleContext,
) -> None:
    """It should get module's type."""
    decoy.when(mock_core.MODULE_TYPE).then_return(ModuleType("heaterShakerModuleType"))
    result = subject.type
    assert result == "heaterShakerModuleType"
